import socket, select
import time
from dataclasses import dataclass

CHERRY_PORT = 10019
CHUNK_MAX_SIZE = 65536

@dataclass
class CherrySocketWrapper:

    def __init__(self, ip, port = CHERRY_PORT):
        self.socket = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        self.socket.connect((ip, port))
        self.ip = ip
        self.port = port
        self.socket.setblocking(False)

    def read_available_bytes(self):
        buffer_end = False
        buffer = b''
        while not buffer_end:
            try:
                chunk = self.socket.recv(CHUNK_MAX_SIZE)
                buffer += chunk
                if chunk.__len__() < CHUNK_MAX_SIZE:
                    buffer_end = True
            except BlockingIOError:
                buffer_end = True
        return buffer

    def write(self, byte_array):
        self.socket.send(byte_array)
        _, writable, _ = select.select([], [self.socket], [])
        while writable.__len__()<=0:
            _, writable, _ = select.select([], [self.socket], [])

    def close(self):
        self.socket.close()




__all__ = ['ReadResponse', 'ReadError']

from ayradb.rest.http.response import Response
from ayradb.core.manage_record.unescape import *
import re
from dataclasses import dataclass

UNESCAPE_ERR_MSG = "Error during unescaping"
SEPARATOR = b'\x3b'


class ReadError:
    # Request errors (local)
    INVALID_FIELD_NAME=-1
    # AyraDB errors
    TABLE_NOT_FOUND=1
    RECORD_NOT_FOUND=2
    INTERNAL_ERROR=100


ERROR_DICT = {
    "ERROR: [00405]:(00008): TFA_EEPA_019: No parameter group found implementing the requested action: probably action group value":ReadError.TABLE_NOT_FOUND,
    "ERROR: [00002]:(00002): TFDRC_007: No data found for this search": ReadError.RECORD_NOT_FOUND,
    # Errors that shouldn't happen since they are internally managed
    "ERROR: [00016]:(00002): TFA_EEPA_018: endpoint action can't be done with this request method": ReadError.INTERNAL_ERROR,
}


@dataclass
class ReadResponse:

    success: bool
    content: dict
    error_code: int
    _error_msg: str

    def __init__(self, success, error_code, content=None, _error_msg=""):
        self.success = success
        self.error_code = error_code
        self.content = content
        self._error_msg = _error_msg

    @staticmethod
    def from_http_response(res: Response):
        error_code = 0
        _error_msg = ""
        content = {}
        if res.status_code == 200:
            success = True
            try:
                body: bytes=res.body
                splitted_body=body.split(SEPARATOR)
                # Parse body
                for cursor in range(0,splitted_body.__len__(),2):
                    # splitted body is organized as [key, value, key, value,...]
                    field_key=splitted_body[cursor]
                    field_value=unescape(splitted_body[cursor+1])
                    content[field_key.decode('utf-8')]=field_value

            except UnescapeException:
                success=False
                error_code=ReadError.INTERNAL_ERROR
                _error_msg=UNESCAPE_ERR_MSG
        else:
            # Case error returned by ayra
            success=False
            error=res.body.decode('ascii')
            # Search for server name inside error message
            match = re.search(r'^(([\s\S])+(?=(ERROR)))', error)
            if match is not None:
                # Case server name present in error
                error = error[match.end():]  # Remove server name from error
            internal_error = ReadError.INTERNAL_ERROR
            try:
                error_code = ERROR_DICT[error]
            except KeyError:
                error_code = internal_error
            if error_code == internal_error:
                _error_msg = error
                # TODO: save in a log file
        return ReadResponse(success,error_code,content=content,_error_msg=_error_msg)


from ayradb.core.manage_record.check_existence.response import ContainsResponse
from ayradb.core.manage_record.format_key import format_key
from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.core.endpoints.endpoints import TABLES_ENDPOINT


def check_record_existence(table_name, key):
    # Create query
    query = {
        "action": "table_read_row",
        "table_name": table_name,
        "key_column_name": "key_column",
        "key_value": format_key(key),
        "fields": "**"
    }
    contains_request = Request(Request.METHOD_GET, path=TABLES_ENDPOINT, query=query)
    return RestManager().submit(contains_request, ContainsResponse.from_http_response)


from dataclasses import dataclass
from ayradb.rest.http.header import Header


@dataclass
class Request:

    method: str
    #host: str
    path: str
    query: dict
    headers: dict
    body: bytes

    METHOD_GET = "GET"
    METHOD_POST = "POST"
    METHOD_PUT = "PUT"
    METHOD_DELETE = "DELETE"

    HTTP_PROTO_VERSION = "HTTP/1.1"

    def __init__(self, method, url="", path="", query=None, headers=None, body=b''):
        if headers is None:
            headers = {}
        if query is None:
            query = {}
        self.method = method
        if url != "":
            try:
                url_to_split = url
                try:
                    # Try removing protocol from url
                    url_to_split = url_to_split.split("://")[1]
                except IndexError:
                    # Case relative url (without proto + host)
                    pass
                url_split=url_to_split.split("/")
                self.path = ""
                # Extract path from url:
                i = 1
                while i < url_split.__len__():
                    # TODO: verify partial urls are parsed correctly
                    self.path += "/"+url_split[i]
                    i += 1
                # Extract query from url
                query_split = url_split[url_split.__len__()-1].split("?")
                if query_split.__len__() > 1:
                    self.query = query
            except IndexError:
                raise InvalidUrl("Invalid url format")
        self.path = path
        self.query = query
        self.headers = headers
        self.body = body
        self.headers[Header.CONTENT_LENGTH] = body.__len__()

    def upsert_header(self, key: str, value: str):
        self.headers[key] = value

    def to_byte_array(self):
        query = ""
        index = 0
        for query_param, param_value in self.query.items():
            query += ("?" if index == 0 else "&")+query_param+"="+str(param_value)
            index += 1
        request_string = self.method+" "+self.path+query+" "+Request.HTTP_PROTO_VERSION+"\r\n"
        for header, h_value in self.headers.items():
            request_string += header+": "+str(h_value)+"\r\n"
        request_string += "\r\n"
        request = request_string.encode('utf8')
        request += self.body
        return request


class InvalidUrl(BaseException):
    def __init__(self, m):
        self.message = m

    def __str__(self):
        return self.message



import json

from ayradb.core.manage_table.get_structure.response import GetStructureResponse
from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.core.endpoints.endpoints import SYSTEM_ENDPOINT

QUERY_ACTION = "set_get_system_parameter"
BODY_ACTION = "get"


def get_table_structure(table_name):
    query = {
        "action": QUERY_ACTION,
    }
    body = {
        "action": BODY_ACTION,
        "target_parameter": "table_record_structure",
        "table_name": table_name
    }

    get_table_request = Request(Request.METHOD_POST, path=SYSTEM_ENDPOINT, query=query, body=json.dumps(body).encode('ascii'))
    return RestManager().submit(get_table_request, GetStructureResponse.from_http_response)



from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.rest.promise import Promise
from ayradb.core.endpoints.endpoints import TABLES_ENDPOINT
from ayradb.core.manage_record.format_key import format_key
from ayradb.core.manage_record.read.response import ReadResponse, ReadError
from ayradb.core.validate import validate_field


def read_record(table_name, key, fields=None):
    # Create query
    query = {
        "action": "table_read_row",
        "table_name": table_name,
        "key_column_name": "key_column",
        "key_value": format_key(key),   # Key is escaped
    }
    if fields is None or fields.__len__() == 0:
        # Case read all fields
        query["fields"] = "*"
        # Build request
        read_request = Request(Request.METHOD_GET, path=TABLES_ENDPOINT, query=query)
        # Submit request to rest manager
        return RestManager().submit(read_request, ReadResponse.from_http_response)
    else:
        # Case read only specific fields
        concat= fields[0]
        fields_valid = validate_field(concat)
        for fieldIdx in range(1, fields.__len__()):
            # Concat remaining fields
            curr_field= fields[fieldIdx]
            fields_valid= fields_valid and validate_field(curr_field)
            concat = f'{concat},{curr_field}'
        query["fields"]= concat

        # To reduce number of controls [and faster than if] for valid 
        # queries (most times) we control validity only once after 
        # concatenating fields
        if fields_valid:
            # All fields ok
            # Build request
            read_request = Request(Request.METHOD_GET, path=TABLES_ENDPOINT, query=query)
            # Submit request to rest manager
            return RestManager().submit(read_request, ReadResponse.from_http_response)
        else:
            # At least one invalid field
            return Promise.reject(ReadResponse(False, ReadError.INVALID_FIELD_NAME))  # Solved promise with error


from dataclasses import dataclass
from ayradb.rest.http.header import Header


@dataclass
class Response:

    status_code: int
    status_message: str
    headers: dict
    body: bytes

    # Position trackers
    header_buffer_seek_pos: int
    body_parsing_start_pos: int
    

    #Parsing flags
    flag_proto_ok: bool
    flag_status_code_ok: bool
    flag_status_msg_ok: bool
    flag_headers_ok: bool


    _HTTP_PROTO_BYTES_SIZE=8
    _HTTP_STATUS_CODE_FIRST_BYTE_IDX=9
    _HTTP_STATUS_MSG_FIRST_BYTE_IDX=13
    _HTTP_FIRST_LINE_SIZE=14
    _HTTP_LINE_TERMINATOR=b'\x0d\x0a'

    _HTTP_STATUS_DICT={
        # 1xx INFORMATIONAL
        b'\x31\x30\x30':100, #100 - Continue
        b'\x31\x30\x31':101, #101 - Switching protocols
        b'\x31\x30\x32':102, #102 - Processing
        b'\x31\x30\x33':103, #103 - Early hints
        b'\x31\x39\x38':198, #198 - Cherrydata-provisional-update-servers-coord
        # 2xx SUCCESS
        b'\x32\x30\x30':200, #200 - OK
        b'\x32\x30\x31':201, #201 - Created
        b'\x32\x30\x32':202, #202 - Accepted
        b'\x32\x30\x33':203, #203 - Non-Authoritative Information
        b'\x32\x30\x34':204, #204 - No Content
        b'\x32\x30\x35':205, #205 - Reset Content
        b'\x32\x30\x36':206, #206 - Partial Content
        b'\x32\x30\x37':207, #207 - Multi status
        # 3xx REDIRECT
        # TODO: add missing
        b'\x33\x30\x30':300, #300 - Multiple choices
        b'\x33\x30\x31':301, #301 - Moved permanently
        # 4xx CLIENT ERRORs
        b'\x34\x30\x30':400, #400 - Bad request
        b'\x34\x30\x31':401, #401 - Unauthorized
        b'\x34\x30\x32':402, #402 - Payment required
        b'\x34\x30\x33':403, #403 - Forbidden
        b'\x34\x30\x34':404, #404 - Not found
        b'\x34\x30\x35':405, #405 - Method not allowed
        # 5xx CLIENT ERRORs
        b'\x35\x30\x30':500, #500 - Internal Server Error
        b'\x35\x30\x31':501, #501 - Not Implemented
        b'\x35\x30\x32':502, #502 - Bad Gateway
        b'\x35\x30\x33':503, #503 - Service Unavailable
        b'\x35\x30\x34':504, #504 - Gateway Timeout
        b'\x35\x30\x35':505, #505 - HTTP Version Not Supported
    }

    def __init__(self):
        self.flag_proto_ok=False
        self.flag_status_code_ok=False
        self.flag_status_msg_ok=False
        self.flag_headers_ok=False
        self.header_buffer_seek_pos=0
        self.body_parsing_start_pos=0
        self.headers={}
        self.body=b''

    
    def from_byte_array(self, byte_array):
        if not self.flag_proto_ok:
            # HTTP proto still not parsed
            if byte_array.__len__()<Response._HTTP_PROTO_BYTES_SIZE:
                # Not enough bytes to check protocol
                return 0
            else:
                # Verify that in first 8 bytes there is HTTP/1.1
                self.flag_proto_ok=byte_array[0:Response._HTTP_PROTO_BYTES_SIZE]==b'\x48\x54\x54\x50\x2F\x31\x2e\x31'
                if not self.flag_proto_ok:
                    # Malformed message: HTTP/1.1 isn't in first 8 bytes
                    return -1

        if not self.flag_status_code_ok:
            # Status code not parsed

            # Status code still not parsed
            if byte_array.__len__()<Response._HTTP_FIRST_LINE_SIZE:
                # Not enough bytes to check status
                return 0
            # Check that a space is present after HTTP/1.1
            if byte_array[Response._HTTP_PROTO_BYTES_SIZE] != 0x20:
                # Case space not present (malformed message)
                return -1
            # Check http status code
            status_bytes=byte_array[Response._HTTP_STATUS_CODE_FIRST_BYTE_IDX:Response._HTTP_STATUS_CODE_FIRST_BYTE_IDX+3]
            self.status_code = Response._HTTP_STATUS_DICT.get(status_bytes)
            if self.status_code is None:
                # Bytes didn't correspond to a valid HTTP status
                return -1
            # Check that a space is present after the status
            if byte_array[Response._HTTP_STATUS_CODE_FIRST_BYTE_IDX+3] != 0x20:
                # Space not present (malformed message)
                return -1
            #Status code parsed correctly
            self.flag_status_code_ok=True

        if not self.flag_status_msg_ok:
            # Status message not parsed

            # Start parsing from status message first byte position
            status_msg_pos= Response._HTTP_STATUS_MSG_FIRST_BYTE_IDX
            # Seek next \r\n
            next_rn_position=Response._seek_line_terminator(byte_array,start=status_msg_pos)
            if next_rn_position==-1:
                # \r\n not found (not enough bytes)
                return 0
            else:
                self.flag_status_msg_ok=True
                # Decode status msg
                self.status_message=byte_array[status_msg_pos:next_rn_position]\
                                        .decode('utf8')
                # Start searching for the headers from the first byte
                # after the line terminator
                self.header_buffer_seek_pos=next_rn_position+Response._HTTP_LINE_TERMINATOR.__len__()
                # TODO save current parser position
        while not self.flag_headers_ok:
            # Header still not parsed

            # Seek next \r\n starting from the first not parsed
            # byte of the headers
            next_rn_position=Response._seek_line_terminator(byte_array,start=self.header_buffer_seek_pos)
            if next_rn_position==-1:
                # \r\n not found (not enough bytes)
                return 0
            if next_rn_position==self.header_buffer_seek_pos:
                # This \r\n is preceded by another \r\n
                # Double line terminator means headers end
                # Since a HTTP message has at least 1 header (Content-Length)
                # we can skip the case with 0 headers
                self.flag_headers_ok=True
                self.body_parsing_start_pos=self.header_buffer_seek_pos+Response._HTTP_LINE_TERMINATOR.__len__()
                break
            # Decode header row
            header_row=byte_array[self.header_buffer_seek_pos:next_rn_position]\
                            .decode('utf8')
            # Split header in key value
            splitted_header=header_row.split(':')
            if splitted_header[1][0]==' ':
                # Remove whitespace before adding key value pair to headers
                self.headers[splitted_header[0]]=splitted_header[1][1:]
            else:
                # Add key value pair to headers
                self.headers[splitted_header[0]]=splitted_header[1]
            # Move header start position after \r\n
            self.header_buffer_seek_pos=next_rn_position+Response._HTTP_LINE_TERMINATOR.__len__()
        
        # Get message length
        content_length = int(self.headers.get(Header.CONTENT_LENGTH))
        missing_bytes = content_length - self.body.__len__()
        parsable_bytes=0
        if byte_array.__len__()<self.body_parsing_start_pos+missing_bytes:
            # Case not enough bytes to parse the whole body
            # Parse all the available bytes
            parsable_bytes = byte_array.__len__()-self.body_parsing_start_pos
        else:
            parsable_bytes = missing_bytes
        self.body+=byte_array[self.body_parsing_start_pos:self.body_parsing_start_pos+parsable_bytes]
        self.body_parsing_start_pos=self.body_parsing_start_pos+parsable_bytes
        if self.body.__len__()!=content_length:
            # Missing bytes from body
            return 0
        else:
            # Body fully parsed
            return self.body_parsing_start_pos

    @staticmethod
    def _seek_line_terminator(byte_array,start=0):
        return Response._seek_sub_slice(byte_array, Response._HTTP_LINE_TERMINATOR,start=start)
                
    @staticmethod
    def _seek_sub_slice(byte_array,sub_slice,start=0):
        #Start searching from the start position
        idx=start
        sub_sl_len=sub_slice.__len__()
        # Search up for the first sub slice byte 
        # up to (byte_array.length - sub_slice.length) byte
        # After that we don't have enough byte to find the pattern
        while idx+sub_sl_len-1<byte_array.__len__():
            if byte_array[idx] == sub_slice[0]:
                # First byte of the pattern found
                idx+=1
                # Move to next and check remaining bytes
                if byte_array[idx:idx+sub_sl_len-1]==sub_slice[1:]:
                    #Pattern found (Move back to original position)
                    return idx-1
            else:
                idx+=1
        # Pattern not found
        return -1
                
                        
       


from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.rest.promise import Promise
from ayradb.core.endpoints.endpoints import TABLES_ENDPOINT
from ayradb.core.manage_record.format_key import format_key
from ayradb.core.manage_record.delete.response import DeleteResponse, DeleteError
from ayradb.core.validate import validate_field


def delete_record(table_name, key, fields=None):
    # Create query
    query = {
        "action": "table_delete_row",
        "table_name": table_name,
        "key_column_name": "key_column",
        "key_value": format_key(key),  # Key is escaped
    }
    are_fields_valid = True
    if fields is not None and fields.__len__() > 0:
        # Case delete only specific fields [NO_SQL table only]
        concat = fields[0]
        fields_valid = validate_field(concat)
        for fieldIdx in range(1, fields.__len__()):
            # Concat remaining fields
            curr_field = fields[fieldIdx]
            fields_valid = fields_valid and validate_field(curr_field)
            concat = f'{concat},{curr_field}'
        query["fields"] = concat
    if not are_fields_valid:
        return Promise.reject(DeleteResponse(False, DeleteError.INVALID_FIELD_NAME))
    delete_request = Request(Request.METHOD_DELETE, path=TABLES_ENDPOINT, query=query)
    return RestManager().submit(delete_request, DeleteResponse.from_http_response)



__all__ = ['GetStructureResponse', 'GetStructureError']

from ayradb.rest.http.response import Response
import re
import json
from dataclasses import dataclass
from ayradb.core.manage_table.column import Column


class GetStructureError:
    # Request errors (local)
    INVALID_FIELD_NAME = -1
    # AyraDB errors
    TABLE_NOT_FOUND = 1
    INTERNAL_ERROR = 100


ERROR_DICT = {
    "ERROR: [00005]:(00002): QFISGSPTRS_005: table not found":GetStructureError.TABLE_NOT_FOUND,
    # Errors that shouldn't happen since they are internally managed
    "ERROR: [00016]:(00002): TFA_EEPA_018: endpoint action can't be done with this request method": GetStructureError.INTERNAL_ERROR,
    "ERROR: [00019]:(00001): TFA_EEPA_019.004: Too many parameters in the request uri": GetStructureError.INTERNAL_ERROR
}


@dataclass
class GetStructureResponse:

    success: bool
    error_code: int
    _error_msg: str
    structure: list

    def __init__(self, success, error_code,_error_msg="", description=None):
        self.success = success
        self.error_code = error_code
        self._error_msg = _error_msg
        self.structure = description

    @staticmethod
    def from_http_response(res: Response):
        error_code = 0
        _error_msg = ""
        fields = []
        if res.status_code == 200:
            success = True
            body = json.loads(res.body)
            body_description = body["column_descriptions"]
            print(body_description)
            for idx in range(0, body_description.__len__()):
                fields.append({Column.NAME: body_description[idx]["column_label"]})
                if "column_max_net_length" in body_description[idx] and body_description[idx]["column_max_net_length"] is not None:
                    fields[idx][Column.MAX_LENGTH] = body_description[idx]["column_max_net_length"] / 2  # FIXME: Remove /2 if escape factor is enabled
        else:
            # Case error returned by ayra
            success=False
            error=res.body.decode('ascii')
            # Search for server name inside error message
            match = re.search(r'^(([\s\S])+(?=(ERROR)))', error)
            if match is not None:
                # Case server name present in error
                error = error[match.end():]  # Remove server name from error
            internal_error = GetStructureError.INTERNAL_ERROR
            try:
                error_code = ERROR_DICT[error]
            except KeyError:
                error_code = internal_error
            if error_code == internal_error:
                _error_msg = error
                # TODO: save in a log file

        return GetStructureResponse(success,error_code,_error_msg=_error_msg, description=fields)



__all__ = ['ScanResponse', 'ScanError']

from ayradb.rest.http.response import Response
from ayradb.core.manage_record.unescape import *
import re
from dataclasses import dataclass

SEPARATOR = b'\x3b'
LAST_HASH_KEY_NAME = b'__scan_table__last_key_hash__'
KEY_COLUMN_NAME = b'key_column'


class ScanError:
    # Request errors (local)
    INVALID_FIELD_NAME=-1
    # AyraDB errors
    TABLE_NOT_FOUND=1
    SEGMENT_END=2
    INVALID_SEGMENT = 3
    SEGMENT_OUT_OF_RANGE = 4
    INTERNAL_ERROR=100


ERROR_DICT = {
    "ERROR: [00405]:(00008): TFA_EEPA_019: No parameter group found implementing the requested action: probably action group value":ScanError.TABLE_NOT_FOUND,
    "ERROR: [00002]:(00002): TFDRC_002.003: scan_table: no next keys found": ScanError.SEGMENT_END,
    "ERROR: [00019]:(00001): TFA_MTKOSCANTABLE_012: message_reference->body: segment: not a number": ScanError.INVALID_SEGMENT,
    "ERROR: [00019]:(00001): TFA_MTKOSCANTABLE_013: message_reference->body: segment: out of range": ScanError.SEGMENT_OUT_OF_RANGE,
    # Errors that shouldn't happen since they are internally managed
    "ERROR: [00016]:(00002): TFA_EEPA_018: endpoint action can't be done with this request method": ScanError.INTERNAL_ERROR,
}


@dataclass
class ScanResponse:

    success: bool
    content: dict
    last_hash: str
    key_column: str
    segment_ended: bool
    error_code: int
    _error_msg: str

    def __init__(self, success, error_code, last_hash, key_column, segment_end=False, content=None, _error_msg=""):
        self.success = success
        self.error_code = error_code
        self.last_hash = last_hash
        self.key_column = key_column
        self.segment_ended = segment_end
        self.content = content
        self._error_msg = _error_msg

    @staticmethod
    def from_http_response(res: Response):
        error_code = 0
        _error_msg = ""
        last_hash = ""
        key_column = ""
        content = {}
        if res.status_code == 200:
            success = True
            try:
                body: bytes = res.body
                splitted_body = body.split(SEPARATOR)
                # Parse body
                for cursor in range(0, splitted_body .__len__(),2):
                    # splitted body is organized as [key, value, key, value,...]
                    field_key = splitted_body[cursor]
                    field_value = unescape(splitted_body[cursor+1])
                    if field_key == LAST_HASH_KEY_NAME:
                        # Case last key hash found
                        last_hash = field_value.decode('ascii')
                    elif field_key == KEY_COLUMN_NAME:
                        # Case key column field found
                        key_column = field_value.decode('ascii')
                    else:
                        # Case record field found
                        field_value = unescape(field_value)
                        content[field_key.decode('utf-8')] = field_value

            except UnescapeException:
                success = False
                error_code = ScanError.INTERNAL_ERROR
                _error_msg = "Parsing error"
        else:
            # Case error returned by ayra
            success = False
            error = res.body.decode('ascii')
            # Search for server name inside error message
            match = re.search(r'^(([\s\S])+(?=(ERROR)))', error)
            if match is not None:
                # Case server name present in error
                error = error[match.end():]  # Remove server name from error
            internal_error = ScanError.INTERNAL_ERROR
            try:
                error_code = ERROR_DICT[error]
            except KeyError:
                error_code = internal_error
            if error_code == internal_error:
                _error_msg = error
                # TODO: save in a log file
        return ScanResponse(success, error_code, last_hash, key_column, content=content, _error_msg=_error_msg)



__all__ = ['Promise']

import time
from dataclasses import dataclass
from threading import Lock

from ayradb.rest.http.request import Request
from ayradb.rest.http.response import Response

WAIT_SLEEP = 0.00001 #10us

@dataclass
class Promise:

    __id_count = 0
    __lock = Lock()

    _req: Request
    res_method: type(lambda x:Response)
    informational_msgs: []
    lock: Lock

    def __init__(self,request,build:type(lambda x:Response)):
        self._req=request
        Promise.__lock.acquire()
        self.cID = Promise.__id_count
        Promise.__id_count+=1
        Promise.__lock.release()
        self._res = None
        self.res_method=build
        self.informational_msgs=[]
        self.lock=Lock()

    def get_request(self):
        return self._req
    
    def wait_response(self):
        self.lock.acquire()
        while self._res is None:
            self.lock.release()
            time.sleep(WAIT_SLEEP)
            self.lock.acquire()
        self.lock.release()
        return self._res

    def submit(self,response):
        self.lock.acquire()
        self._res = self.res_method(response)
        self.lock.release()

    def submit_informational(self,response):
        self.informational_msgs.append(response)

    @staticmethod
    def reject(error_object):
        promise = Promise(None, None)
        promise._res=error_object
        return promise



__all__ = ['RestManager']

import time
import threading
from dataclasses import dataclass
from queue import Queue, Empty
from threading import Thread, Lock, Condition

from ayradb.rest.http.response import Response, Header
from ayradb.rest.promise import Promise
from ayradb.rest.socket.wrapper import CherrySocketWrapper
from ayradb.utils.singleton import Singleton
from ayradb.core.keep_alive.request import generate_keep_alive_promise

DISCONNECTION_TIME = 5  # 10 seconds -> keep 2 second margin
MAX_CONN = 1
MAX_CONN_ON_NODE = 1
SLEEP_ADDING_CONNECTION = 0.00005  # 50us
SLEEP_EMPTY_QUEUE_SEC = 0.00001  # 10us
SLEEP_SHUTDOWN_CONTROLLER = 0.0001  # 100us
SLEEP_TCP_BUFFER_EMPTY = 0.00001  # 10us
MIN_NON_INFO_STATUS_CODE = 200

@dataclass
class RestManager(metaclass=Singleton):

    lock: Lock
    condition: Condition
    process: bool
    disconnect: bool
    conn_on_node: dict
    connections: int
    connections_dict: dict
    promises_list: list

    def __init__(self):
        self.process = False
        self.disconnect = False
        self.lock = Lock()
        self.updating_runtime_data = False
        self.connections=0
        self.conn_on_node={}
        self.connections_dict = {}
        self.promises_lists_dict = {}
        self.lock_dict = {}
        self.condition = Condition(lock=self.lock)
        self.conn_on_node = {}
    
    def connect(self, ip, port = None):
        self.lock.acquire()
        connections = self.conn_on_node.get(ip)
        if self.connections<MAX_CONN and\
            (connections is None or connections<MAX_CONN_ON_NODE):
            # A new connection can be established
            socket = None
            if port is None:
                socket = CherrySocketWrapper(ip)
            else:
                socket = CherrySocketWrapper(ip, port = port)
            socket_id = self.connections
            self.connections_dict[socket_id]=socket
            self.promises_lists_dict[socket_id] = []
            self.lock_dict[socket_id] = Lock()
            self.connections += 1
            # Update connection on node counter
            if connections is None:
                # Add node to connection list
                self.conn_on_node[ip] = 1
            else: 
                # Increment node connections
                self.conn_on_node[ip] = self.conn_on_node[ip]+1
            self.lock.release()
            curr_th = Thread(target=self._process_data_exchange, args=(socket_id,))
            curr_th.start()
            with self.condition:
                while not self.process:
                    # Wait for thread connection setup
                    self.condition.wait()
        else:
            self.lock.release()

    def is_connected(self):
        self.lock.acquire()
        is_processing = self.process
        self.lock.release()
        return is_processing
    
    def stop_processing(self):
        self.lock.acquire()
        self.disconnect = True
        self.lock.release()
        th = Thread(target=self.__manage_stop)
        th.start()
        

    def __manage_stop(self):
        self.lock.acquire()
        while self.connections>0:
            self.lock.release()
            time.sleep(SLEEP_SHUTDOWN_CONTROLLER)
            self.lock.acquire()
        self.process = False
        self.lock.release()

    def _process_data_exchange(self,socket_id):
        socket: CherrySocketWrapper = self.connections_dict[socket_id]
        promises_list: list = self.promises_lists_dict[socket_id]
        lock: Lock = self.lock_dict[socket_id]
        # Last msg received time (initialize here since connection is open)
        last_rcv_time = time.time()
        # Create buffer for response bytes
        res_buffer = b''
        # Start processing incoming requests
        continue_processing = True
        with self.condition:
            self.process = continue_processing
            # Notify main thread that socket thread is
            # ready to process requests
            self.condition.notify()
        while not self.disconnect:
            response_parsed = False
            res = Response()
            last_read_len = 0
            promise = None
            is_promise_keep_alive = False
            while not response_parsed:
                # Parse responses related to request sent
                res_buffer += socket.read_available_bytes()
                curr_read_len = res_buffer.__len__()
                if curr_read_len == last_read_len:
                    # No new bytes
                    if res_buffer.__len__() == 0:
                        # Case nothing received from ayra
                        if time.time()-last_rcv_time > DISCONNECTION_TIME and not is_promise_keep_alive:
                            lock.acquire()
                            is_promise_keep_alive = True
                            promise = generate_keep_alive_promise()
                            req = promise.get_request()
                            req.upsert_header(Header.HOST, socket.ip)
                            socket.write(req.to_byte_array())
                            lock.release()
                        if self.disconnect:
                            break
                    # Sleep
                    time.sleep(SLEEP_TCP_BUFFER_EMPTY)
                else:
                    last_read_len = curr_read_len
                    last_rcv_time = time.time()
                bytes_read = res.from_byte_array(res_buffer)
                if bytes_read > 0:
                    # Response found in buffer
                    if last_read_len > bytes_read:
                        res_buffer = res_buffer[bytes_read:last_read_len]
                    else:
                        res_buffer = b''
                    # Update buffer length of current read
                    last_read_len = res_buffer.__len__()
                    try:
                        if promise is None:
                            lock.acquire()
                            promise = promises_list.pop(0)
                            lock.release()
                        if res.status_code >= MIN_NON_INFO_STATUS_CODE:
                            # Case non informational response [>=200]
                            promise.submit(res)
                            response_parsed = True
                            promise = None
                            is_promise_keep_alive = False
                        else:
                            # Case informational response [1xx]
                            promise.submit_informational(res)
                            res = Response()
                    except IndexError:
                        # Informational not paired to a request
                        # Ignore and get ready for another response
                        res = Response()
                        lock.release()
                    
            
        self.lock.acquire()
        self.connections -= 1
        self.conn_on_node[socket.ip] -= 1
        socket.close()
        self.lock.release()
        return True

    def submit(self, request, build:type(lambda x:Response)):
        # Create promise and set promise id in header
        promise = Promise(request,build)
        # TODO use dict len to block insertion
        promise_id = promise.cID
        # Select a connection according to promise id
        connection_id = promise_id%self.connections
        socket: CherrySocketWrapper = self.connections_dict[connection_id]
        promise_list: list = self.promises_lists_dict[connection_id]
        lock: Lock = self.lock_dict[connection_id]
        # Add missing headers
        request.upsert_header(Header.CONNECTION, "Keep-Alive")
        request.upsert_header(Header.HOST, socket.ip)
        lock.acquire()
        while promise_list.__len__()>=32:
            lock.release()
            time.sleep(SLEEP_EMPTY_QUEUE_SEC)
            lock.acquire()
        promise_list.append(promise)
        socket.write(request.to_byte_array())
        lock.release()
        return promise


ESCAPE_BYTE=b'\x5c'
DOUBLE_ESCAPE_BYTE=b'\x5c\x5c'

UNESCAPE_BYTE_DICT={
    0x7A: b'\x00', #null 	\z 0x5c 0x7a
    0x71: b'\x22', #" \q 0x5c 0x71 
    0x73: b'\x3B', #; \s 0x5c 0x73 
    0x5C: b'\x5C', #\ \\ 0x5c 0x5c
    0x6E: b'\x0A', #line-feed \n 0x5c 0x6e 
    0x66: b'\x0C', #form-feed \f 0x5c 0x66
    0x63: b'\x0D', #carriage-return \c 0x5c 0x63 
}

def unescape(string:bytes):
    str_len=string.__len__()
    idx=0
    splitted_response = string.split(DOUBLE_ESCAPE_BYTE)
    # Unescape splitted response [0]
    current_split=splitted_response[0].split(ESCAPE_BYTE)
    splitted_response[0]=current_split[0]
    for idx in range(1,current_split.__len__()):
        # For every sub array of splitted_response[0] splitted at \
        to_unescape = current_split[idx][0]
        unescaped_byte = UNESCAPE_BYTE_DICT.get(to_unescape)
        if unescaped_byte is None:
            # Error in escaping or unescaping (unexpected char)
            raise UnescapeException("Unescape char not found") 
        # Modify byte to unescape
        splitted_response[0]+=unescaped_byte+current_split[idx][1:]
    # Unescape splitted response [1:]
    for cursor in range(1,splitted_response.__len__()):
        # For each array splitted at \\
        current_split=splitted_response[cursor].split(ESCAPE_BYTE)
        splitted_response[cursor]=current_split[0]
        for idx in range(1,current_split.__len__()):
            # For each sub array splitted at \
            to_unescape = current_split[idx][0]
            unescaped_byte = UNESCAPE_BYTE_DICT.get(to_unescape)
            if unescaped_byte is None:
                # Error in escaping or unescaping (unexpected char)
                raise UnescapeException("Unescape char not found") 
            # Modify byte to unescape
            splitted_response[cursor]+=unescaped_byte+current_split[idx][1:]
        # Add unescaped slash (double slash in escaped version)
        splitted_response[0]+=b'\x5c'+splitted_response[cursor]
    return splitted_response[0]


# UNESCAPE_BYTE_DICT={
#     b'\x7A': b'\x00', #null 	\z 0x5c 0x7a
#     b'\x71': b'\x22', #" \q 0x5c 0x71 
#     b'\x73': b'\x3B', #; \s 0x5c 0x73 
#     b'\x5C': b'\x5C', #\ \\ 0x5c 0x5c
#     b'\x6E': b'\x0A', #line-feed \n 0x5c 0x6e 
#     b'\x66': b'\x0C', #form-feed \f 0x5c 0x66
#     b'\x63': b'\x0D', #carriage-return \c 0x5c 0x63 
# }
# def unescape(string:bytes):
#     str_len=string.__len__()
#     idx=0
#     escaped_field=bytearray() # Unescaped max size = escaped size
#     while idx<str_len:
#         curr_byte=string[idx:idx+1]
#         if curr_byte==ESCAPE_BYTE:
#             # Case UNESCAPE
#             # Check next byte
#             idx+=1
#             escape = UNESCAPE_BYTE_DICT.get(string[idx:idx+1])
#             if escape is None:
#                 # Error in escaping or unescaping (unexpected char)
#                 raise UnescapeException("Unescape char not found") 
#             escaped_field+=escape
#         else:
#             # Case DON'T unescape
#             escaped_field+=curr_byte
#         # Move to string next byte
#         idx+=1
#     return bytes(escaped_field)

# def unescape(string:bytes):
#     str_len=string.__len__()
#     idx=0
#     tot_size=0
#     escaped_field=bytearray(str_len) # Unescaped max size = escaped size
#     while idx<str_len:
#         curr_byte=string[idx:idx+1]
#         if curr_byte==ESCAPE_BYTE:
#             # Case UNESCAPE
#             # Check next byte
#             idx+=1
#             escape = UNESCAPE_BYTE_DICT.get(string[idx:idx+1])
#             if escape is None:
#                 # Error in escaping or unescaping (unexpected char)
#                 raise UnescapeException("Unescape char not found") 
#             escaped_field[tot_size:tot_size+1]=escape
#         else:
#             # Case DON'T unescape
#             escaped_field[tot_size]=string[idx]
#         # Move to string next byte
#         idx+=1
#         # Move to write next unescaped byte 
#         tot_size+=1
#     return bytes(escaped_field[0:tot_size])

class UnescapeException(BaseException):
    def __init__(self, m):
        self.message = m
    def __str__(self):
        return self.message


from ayradb.core.manage_table.column import *
from ayradb.core.manage_table.create_lists import create_server_list_from_list, create_column_list
from ayradb.core.manage_table.types import TABLE_FIXED_LENGTH
from ayradb.rest.promise import Promise
from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.core.endpoints.endpoints import SYSTEM_ENDPOINT, TABLES_ENDPOINT
from ayradb.core.manage_table.restructure.response import RestructureTableResponse, RestructureTableError

import json

ACTION = "table_restructure"
MINIMUM_REPLICATION = 1
DEFAULT_ESCAPE_FACTOR = 1

def restructure_table(name, fill_type, user_columns:[], servers: [], replication, table_max_size, key_max_size, escape_factor=DEFAULT_ESCAPE_FACTOR):

    server_list = create_server_list_from_list(servers)
    query = {
            "action": ACTION
        }

    body = {
        "table_name": name,
        "table_clone_name": "",
        "endpoint": TABLES_ENDPOINT,
        "table_description": "table_description",
        "restructuration_type": "servers",
        "server_list": server_list
    }
    if table_max_size is not None:
        body["new_table_max_size_bytes"] = table_max_size
    if replication is not None:
        if replication > MINIMUM_REPLICATION > servers.__len__():
            return Promise.reject(RestructureTableResponse(False, RestructureTableError.NOT_ENOUGH_NODES))
        body["new_table_replication_factor"] = replication
    if user_columns is not None:
        if fill_type == TABLE_FIXED_LENGTH:
            for idx in range(0, user_columns.__len__()):
                if user_columns[idx].get(COLUMN_MAX_LENGTH) is None:
                    return Promise.reject(RestructureTableResponse(False, RestructureTableError.FIELD_MAX_LENGTH_REQUIRED))
            if key_max_size <= 0:
                return Promise.reject(RestructureTableResponse(False, RestructureTableError.KEY_MAX_LENGTH_REQUIRED))
        column_list = create_column_list(user_columns, fill_type, escape_factor + 1, key_max_size=key_max_size)
        body["new_record_structure"] = column_list

    create_request = Request(Request.METHOD_POST, path=SYSTEM_ENDPOINT, query=query, body=json.dumps(body).encode('ascii'))
    return RestManager().submit(create_request, RestructureTableResponse.from_http_response)



__all__ = ['CreateTableResponse', 'CreateTableError']

from ayradb.rest.http.response import Response
import re
from dataclasses import dataclass


class CreateTableError:
    # Request errors (local)
    FIELD_MAX_LENGTH_REQUIRED = -1  # Also managed remotely
    KEY_MAX_LENGTH_REQUIRED = -2  # Also managed remotely
    NOT_ENOUGH_NODES = -3
    # AyraDB errors
    ALREADY_EXISTS = 1
    INTERNAL_ERROR = 100


ERROR_DICT = {
    "ERROR: [00019]:(00001): QFICRTAB_011: request json: table_description: table_conventional_name: the table already exists": CreateTableError.ALREADY_EXISTS,
    # Errors that should be verified also locally
    "ERROR: [00019]:(00001): CTDFJ_030.001: column_descriptions: column_max_net_length: not found": CreateTableError.FIELD_MAX_LENGTH_REQUIRED,
    "ERROR: [00019]:(00001): CTDFJ_030.002: column_descriptions: column_max_net_length: not a number": CreateTableError.KEY_MAX_LENGTH_REQUIRED,
    # Errors that shouldn't happen since they are internally managed
    "ERROR: [00016]:(00002): TFA_EEPA_018: endpoint action can't be done with this request method": CreateTableError.INTERNAL_ERROR,
}


@dataclass
class CreateTableResponse:

    success: bool
    error_code: int
    _error_msg: str

    def __init__(self, success, error_code,_error_msg=""):
        self.success = success
        self.error_code = error_code
        self._error_msg = _error_msg

    @staticmethod
    def from_http_response(res: Response):
        if res.status_code == 200:
            return CreateTableResponse(True, 0)
        else:
            # Case error returned by ayra
            _error_msg = ""
            error = res.body.decode('ascii')
            # Search for server name inside error message
            match = re.search(r'^(([\s\S])+(?=(ERROR)))', error)
            if match is not None:
                # Case server name present in error
                error = error[match.end():]  # Remove server name from error
            internal_error = CreateTableError.INTERNAL_ERROR
            try:
                error_code = ERROR_DICT[error]
            except KeyError:
                error_code = internal_error
            if error_code == internal_error:
                _error_msg = error
                # TODO: save in a log file

            return CreateTableResponse(False, error_code, _error_msg=_error_msg)



__all__ = ['RestructureTableResponse', 'RestructureTableError']

from ayradb.rest.http.response import Response
import re
from dataclasses import dataclass


class RestructureTableError:    # TODO
    # Request errors (local)
    FIELD_MAX_LENGTH_REQUIRED = -1  # Also managed remotely
    KEY_MAX_LENGTH_REQUIRED = -2  # Also managed remotely
    NOT_ENOUGH_NODES = -3
    # AyraDB errors
    INTERNAL_ERROR = 100

ERROR_DICT = {

}


@dataclass
class RestructureTableResponse:

    success: bool
    error_code: int
    _error_msg: str

    def __init__(self, success, error_code,_error_msg=""):
        self.success = success
        self.error_code = error_code
        self._error_msg = _error_msg

    @staticmethod
    def from_http_response(res: Response):
        if res.status_code == 200:
            return RestructureTableResponse(True, 0)
        else:
            # Case error returned by ayra
            _error_msg = ""
            error = res.body.decode('ascii')
            # Search for server name inside error message
            match = re.search(r'^(([\s\S])+(?=(ERROR)))', error)
            if match is not None:
                # Case server name present in error
                error = error[match.end():]  # Remove server name from error
            internal_error = RestructureTableError.INTERNAL_ERROR
            try:
                error_code = ERROR_DICT[error]
            except KeyError:
                error_code = internal_error
            if error_code == internal_error:
                _error_msg = error
                # TODO: save in a log file

            return RestructureTableResponse(False, error_code, _error_msg=_error_msg)



import re

ESCAPE_DICT = {
    b'\x00': rb'\z',    # null        -> \z 0x5c 0x7a
    b'"': rb'\q',       # "           -> \q 0x5c 0x71
    b';': rb'\s',       # ;           -> \s 0x5c 0x73
    b'\\': rb'\\',      # \           -> \\ 0x5c 0x5c
    b'\n': rb'\n',      # line-feed   -> \n 0x5c 0x6e
    b'\f': rb'\f',      # form-feed   -> \f 0x5c 0x66
    b'\r': rb'\c',      # carr-return -> \c 0x5c 0x63
}

ESCAPE_CLASS = '[' + ''.join(r'\x' + x.hex() for x in ESCAPE_DICT) + ']'
ESCAPE_REGEX = re.compile(ESCAPE_CLASS.encode())


def escape(string: bytes) -> bytes:
    return re.sub(ESCAPE_REGEX, lambda m: ESCAPE_DICT[m.group(0)], string)

# ESCAPE_DICT={
#     0x00: [0x5C,0x7A], # null        -> \z 0x5c 0x7a
# 	0x22: [0x5C,0x71], # "           -> \q 0x5c 0x71
# 	0x3B: [0x5C,0x73], # ;           -> \s 0x5c 0x73
# 	0x5C: [0x5C,0x5C], # \           -> \\ 0x5c 0x5c
# 	0x0A: [0x5C,0x6E], # line-feed   -> \n 0x5c 0x6e
# 	0x0C: [0x5C,0x66], # form-feed   -> \f 0x5c 0x66
# 	0x0D: [0x5C,0x63], # carr-return -> \c 0x5c 0x63
# }

# def escape(string: bytes):
#     str_len=string.__len__()
#     escaped_array=[]
#     for i in range(0,str_len):
#         curr_byte=string[i]
#         escape = ESCAPE_DICT.get(curr_byte)
#         if escape is None:
#             # Don't escape current byte
#             escaped_array.append(curr_byte)
#         else:
#             # Escape current byte
#             escaped_array.extend(escape)
#     return bytes(escaped_array)

# def escape(string: bytes):
#     str_len=string.__len__()
#     #Escaped max size is twice unescaped size
#     escaped_array=bytearray(2*str_len)
#     total_size=0
#     for i in range(0,str_len):
#         curr_byte=string[i:i+1]
#         escape = ESCAPE_DICT.get(curr_byte)
#         if escape is None:
#             # Don't escape current byte
#             escaped_array[total_size:total_size+1]=curr_byte
#             total_size+=1
#         else:
#             # Escape current byte
#             escaped_array[total_size:total_size+2]=escape
#             total_size+=2
#     return bytes(escaped_array[0:total_size])



# KEY column
COLUMN_NAME_KEY = "key_column"
COLUMN_NAME_NOSQL_VALUE = "value"

# Column build labels
COLUMN_NUMBER = "column_number"
COLUMN_LABEL = "column_label"
COLUMN_SQL_TYPE = "column_sql_type"
COLUMN_HTTP_TYPE = "column_http_type"
COLUMN_MAX_LENGTH="column_max_net_length"

# Column labels provided to users
class Column:
    NAME = COLUMN_LABEL
    MAX_LENGTH = COLUMN_MAX_LENGTH

# Colum types
class SQLFieldType:
    TEXT="TEXT"
    

class HTTPFieldType:
    TEXT_HTML='text\\/html'


__all__ = []

from ayradb.core.ayradb import AyraDB
from ayradb.core.manage_table import column
from ayradb.core import table

from ayradb.core.manage_record.scan.response_scan import ScanError
from ayradb.core.manage_record.read.response import ReadError
from ayradb.core.manage_record.upsert.response import UpsertError
from ayradb.core.manage_record.scan.response_scan_init import ScanInitError
from ayradb.core.manage_record.delete.response import DeleteError
from ayradb.core.manage_record.check_existence.response import ContainsError
from ayradb.core.manage_table.delete.response import DeleteTableError
from ayradb.core.manage_table.create.response import CreateTableError
from ayradb.core.manage_table.truncate.response import TruncateTableError
from ayradb.core.manage_table.get_structure.response import GetStructureError

from ayradb.core.manage_record.scan.response_scan import ScanResponse
from ayradb.core.manage_record.read.response import ReadResponse
from ayradb.core.manage_record.upsert.response import UpsertResponse
from ayradb.core.manage_record.scan.response_scan_init import ScanInitResponse
from ayradb.core.manage_record.delete.response import DeleteResponse
from ayradb.core.manage_record.check_existence.response import ContainsResponse
from ayradb.core.manage_table.delete.response import DeleteTableResponse
from ayradb.core.manage_table.create.response import CreateTableResponse
from ayradb.core.manage_table.truncate.response import TruncateTableResponse
from ayradb.core.manage_table.get_structure.response import GetStructureResponse





from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.core.endpoints.endpoints import SYSTEM_ENDPOINT
from ayradb.core.manage_nodes.add_node.response import AddNodeResponse
import json


def add_node(name, ip, hard_disk: []):
    query = {
        "action": "add_servers"
    }

    body = {
        "server_name": name,
        "server_ip_address": ip
    }

    request = Request(Request.METHOD_POST, path=SYSTEM_ENDPOINT, query=query, body=json.dumps(body).encode('ascii'))
    return RestManager().submit(request, AddNodeResponse.from_http_response)



__all__ = ['DeleteTableResponse', 'DeleteTableError']

from ayradb.rest.http.response import Response
import re
from dataclasses import dataclass


class DeleteTableError:
    # Request errors (local)
    INVALID_FIELD_NAME = -1
    # AyraDB errors
    TABLE_NOT_FOUND = 1
    INTERNAL_ERROR = 100


ERROR_DICT = {
    "ERROR: [00005]:(00002): TFA_QDADRTBL_0002: the table does not exist":DeleteTableError.TABLE_NOT_FOUND,
    # Errors that shouldn't happen since they are internally managed
    "ERROR: [00016]:(00002): TFA_EEPA_018: endpoint action can't be done with this request method": DeleteTableError.INTERNAL_ERROR,
}


@dataclass
class DeleteTableResponse:

    _SEPARATOR = b'\x3b'
    success: bool
    error_code: int
    _error_msg: str

    def __init__(self, success, error_code, _error_msg=""):
        self.success = success
        self.error_code = error_code
        self._error_msg = _error_msg

    @staticmethod
    def from_http_response(res: Response):
        if res.status_code == 200:
            return DeleteTableResponse(True, 0)
        else:
            # Case error returned by ayra
            _error_msg = ""
            success = False
            error = res.body.decode('ascii')
            # Search for server name inside error message
            match = re.search(r'^(([\s\S])+(?=(ERROR)))', error)
            if match is not None:
                # Case server name present in error
                error = error[match.end():]  # Remove server name from error
            internal_error = DeleteTableError.INTERNAL_ERROR
            try:
                error_code = ERROR_DICT[error]
            except KeyError:
                error_code = internal_error
            if error_code == internal_error:
                _error_msg = error
                # TODO: save in a log file

            return DeleteTableResponse(success, error_code, _error_msg=_error_msg)



from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.rest.promise import Promise
from ayradb.core.endpoints.endpoints import TABLES_ENDPOINT
from ayradb.core.manage_record.format_key import format_key
from ayradb.core.manage_record.escape import escape
from ayradb.core.manage_record.upsert.response import UpsertResponse, UpsertError
from ayradb.core.validate import validate_field

FIELD_VALUE_SEPARATOR = b';'
FIELD_KEY_SEPARATOR = ","


def upsert_record(table_name: str, key: str, fields: dict) -> Promise:
    # Create query
    query = {
        "action": "table_insert_row",
        "table_name": table_name,
        "key_column_name": "key_column",
        "key_value": format_key(key), # Key is escaped
        "imode":"drop_older_record_with_same_key"
    }
    are_fields_valid = True
    query_fields = ""
    body = bytearray()
    first_el = True
    for field_key, field_value in fields.items():
        # Generate body and fields query argument
        are_fields_valid = are_fields_valid and validate_field(field_key)
        if first_el:
            body += escape(field_value)
            query_fields = field_key
            first_el = False
        else:
            body += FIELD_VALUE_SEPARATOR+escape(field_value)
            query_fields = f'{query_fields}{FIELD_KEY_SEPARATOR}{field_key}'
    if not are_fields_valid:
        return Promise.reject(UpsertResponse(False,UpsertError.INVALID_FIELD_NAME))
    query["fields"] = query_fields
    upsert_request = Request(Request.METHOD_PUT, path=TABLES_ENDPOINT, query=query, body=body)
    return RestManager().submit(upsert_request, UpsertResponse.from_http_response)



from ayradb.core.manage_table.truncate.response import TruncateTableResponse
from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.core.endpoints.endpoints import SYSTEM_ENDPOINT

ACTION = "table_truncate"

def truncate_table(table_name):
    query = {
        "action": ACTION,
        "table_name": table_name
    }

    truncate_table_request = Request(Request.METHOD_POST, path=SYSTEM_ENDPOINT, query=query)
    return RestManager().submit(truncate_table_request, TruncateTableResponse.from_http_response)



__all__ = ['AddNodeResponse', 'AddNodeError']

from ayradb.rest.http.response import Response
import re
from dataclasses import dataclass


class AddNodeError:
    INTERNAL_ERROR = 100

ERROR_DICT = {

}

@dataclass
class AddNodeResponse:
    success: bool
    error_code: int
    _error_msg: str

    def __init__(self, success, error_code, _error_msg=""):
        self.success = success
        self.error_code = error_code
        self._error_msg = _error_msg

    @staticmethod
    def from_http_response(res: Response):
        success = True
        error_code = 0
        _error_msg = ""
        if res.status_code == 200:
            success = True
        else:
            # Case error returned by ayra
            success = False
            error = res.body.decode('ascii')
            # Search for server name inside error message
            match = re.search(r'^(([\s\S])+(?=(ERROR)))', error)
            if match is not None:
                # Case server name present in error
                error = error[match.end():]  # Remove server name from error
            internal_error = AddNodeError.INTERNAL_ERROR
            try:
                error_code = ERROR_DICT[error]
            except KeyError:
                error_code = internal_error
            if error_code == internal_error:
                _error_msg = error
                # TODO: save in a log file

        return AddNodeResponse(success, error_code, _error_msg=_error_msg)





from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.core.endpoints.endpoints import SYSTEM_ENDPOINT
from ayradb.core.manage_nodes.get_servers.response import GetServersResponse
import json


def get_servers():
    query = {
        "action": "set_get_system_parameter"
    }

    body = {
        "target_parameter": "servers_coordinates", 
        "action": "get"
    }

    request = Request(Request.METHOD_POST, path=SYSTEM_ENDPOINT, query=query, body=json.dumps(body).encode('ascii'))
    return RestManager().submit(request, GetServersResponse.from_http_response)


__all__ = ['AyraDB']

import time
from dataclasses import dataclass

from ayradb.core.exceptions.invalid_table import InvalidTableName
from ayradb.core.exceptions.connection import InitalizationError
from ayradb.core.manage_nodes.get_servers.request import get_servers
from ayradb.core.manage_nodes.add_node.request import add_node
from ayradb.core.table import Table
from ayradb.core.validate import validate_table_name
from ayradb.rest.rest_manager import RestManager
from ayradb.rest.promise import Promise

DISCONNECT_SLEEP = 0.1


@dataclass
class AyraDB:

    ip: str
    validated_names: dict

    def __init__(self, ip="", port = None):
        self.validated_names={}
        if ip != "":
            self.ip = ip
            RestManager().connect(ip, port=port)
        elif not RestManager().is_connected():
            raise InitalizationError("Please provide a valid ip on first initialization")

    def table(self, table_name) -> Table:
        is_name_valid = self.validated_names.get(table_name)
        if is_name_valid is None:
            # Validate and cache
            is_name_valid = validate_table_name(table_name)
            self.validated_names[table_name] = is_name_valid
        if not is_name_valid:
            raise InvalidTableName("The inserted table name contains invalid characters. Please respect the following regexp: [a-zA-Z0-9_]")
        return Table(table_name)

    def close_connection(self):
        RestManager().stop_processing()
        while RestManager().is_connected():
            time.sleep(DISCONNECT_SLEEP)

    def get_servers(self) -> dict:
        return get_servers()

    def add_node(self, name, ip, hard_disk: []) -> Promise:
        return add_node(self.name, self.ip)

    def remove_node(self, name, ip):
        pass



__all__ = ['ScanInitResponse', 'ScanInitError']

from ayradb.rest.http.response import Response
import re
from dataclasses import dataclass
import json


class ScanInitError:
    # Request errors (local)
    # AyraDB errors
    TABLE_NOT_FOUND = 1
    INTERNAL_ERROR = 100


ERROR_DICT = {
    "ERROR: TFA_MTKO_015: trans_h__manage_transaction_kick_off__scan_table:ERROR: [00005]:(00002): TFA_MTKOSCANTABLE_008: table not found": ScanInitError.TABLE_NOT_FOUND,
    "ERROR: [00005]:(00002): TFA_MTKOSCANTABLE_008: table not found": ScanInitError.TABLE_NOT_FOUND,
    # Errors that shouldn't happen since they are internally managed
    "ERROR: [00016]:(00002): TFA_EEPA_018: endpoint action can't be done with this request method": ScanInitError.INTERNAL_ERROR,
    "ERROR: TFA_MTKO_015: trans_h__manage_transaction_kick_off__scan_table:ERROR: [00019]:(00001): TFA_MTKOSCANTABLE_008: message_reference->body: subaction: wrong value": ScanInitError.INTERNAL_ERROR,
}


@dataclass
class ScanInitResponse:

    success: bool
    error_code: int
    _error_msg: str
    segments: int

    def __init__(self, success, error_code, n_segments=0, _error_msg=""):
        self.success = success
        self.error_code = error_code
        self.segments = n_segments
        self._error_msg = _error_msg

    @staticmethod
    def from_http_response(res: Response):
        if res.status_code == 200:
            body=json.loads(res.body)
            return ScanInitResponse(True, 0, n_segments=body["n_segments"])
        else:
            # Case error returned by ayra
            _error_msg = ""
            error = res.body.decode('ascii')
            # Search for server name inside error message
            match = re.search(r'^(([\s\S])+(?=(ERROR)))', error)
            if match is not None:
                # Case server name present in error
                error = error[match.end():]  # Remove server name from error
            internal_error = ScanInitError.INTERNAL_ERROR
            try:
                error_code = ERROR_DICT[error]
            except KeyError:
                error_code = internal_error
            if error_code == internal_error:
                _error_msg = error
                # TODO: save in a log file
            return ScanInitResponse(False,error_code,_error_msg=_error_msg)


__all__ = ['TruncateTableResponse', 'TruncateTableError']

from ayradb.rest.http.response import Response
import re
from dataclasses import dataclass


class TruncateTableError:
    # Request errors (local)
    # AyraDB errors
    TABLE_NOT_FOUND = 1
    INTERNAL_ERROR = 100


ERROR_DICT = {
    "ERROR: [00005]:(00002): TFA_QDATRTBL_0002: the table does not exist": TruncateTableError.TABLE_NOT_FOUND,
    # Errors that shouldn't happen since they are internally managed
    "ERROR: [00016]:(00002): TFA_EEPA_018: endpoint action can't be done with this request method": TruncateTableError.INTERNAL_ERROR,
}


@dataclass
class TruncateTableResponse:

    success: bool
    error_code: int
    _error_msg: str

    def __init__(self, success, error_code, _error_msg=""):
        self.success=success
        self.error_code=error_code
        self._error_msg=_error_msg

    @staticmethod
    def from_http_response(res: Response):
        success=True
        error_code=0
        _error_msg=""
        if res.status_code == 200:
            success=True
        else:
            # Case error returned by ayra
            success = False
            error = res.body.decode('ascii')
            # Search for server name inside error message
            match = re.search(r'^(([\s\S])+(?=(ERROR)))', error)
            if match is not None:
                # Case server name present in error
                error = error[match.end():]  # Remove server name from error
            internal_error = TruncateTableError.INTERNAL_ERROR
            try:
                error_code = ERROR_DICT[error]
            except KeyError:
                error_code = internal_error
            if error_code == internal_error:
                _error_msg = error
                # TODO: save in a log file

        return TruncateTableResponse(success, error_code, _error_msg=_error_msg)





__all__ = ['Table']

from dataclasses import dataclass

from ayradb.core.manage_nodes.get_servers.request import get_servers
from ayradb.core.manage_record.check_existence.request import check_record_existence
from ayradb.core.manage_record.delete.request import delete_record
from ayradb.core.manage_record.read.request import read_record
from ayradb.core.manage_record.scan.request_scan import scan
from ayradb.core.manage_record.scan.request_scan_init import scan_init
from ayradb.core.manage_record.upsert.request import upsert_record
from ayradb.core.manage_table.create.request import MINIMUM_REPLICATION, create_table  #
from ayradb.core.manage_table.create.response import CreateTableResponse, CreateTableError
from ayradb.core.manage_table.delete.request import delete_table
from ayradb.core.manage_table.get_structure.request import get_table_structure
from ayradb.core.manage_table.restructure.request import restructure_table
from ayradb.core.manage_table.truncate.request import truncate_table
from ayradb.core.manage_table.types import TABLE_FIXED_LENGTH, TABLE_PADDED, TABLE_NOSQL
from ayradb.rest.promise import Promise


@dataclass
class Table:

    name: str

    TYPE_FIXED_LENGTH = TABLE_FIXED_LENGTH
    TYPE_PADDED = TABLE_PADDED
    TYPE_NOSQL = TABLE_NOSQL

    def __init__(self, table_name: str):
        self.name = table_name

    def read_record(self, key: str, fields: [] = None) -> Promise:
        return read_record(self.name, key, fields=fields)

    def upsert_record(self, key: str, fields: dict) -> Promise:
        return upsert_record(self.name, key, fields)

    def delete_record(self, key: str, fields: [] = None) -> Promise:
        return delete_record(self.name, key, fields=fields)

    def contains_record(self, key: str) -> Promise:
        return check_record_existence(self.name, key)

    def scan_init(self) -> Promise:
        return scan_init(self.name)

    def scan(self, segment: int, fields: [] = None, last_hash: str = None) -> Promise:
        return scan(self.name, segment, fields=fields, last_hash=last_hash)

    def create(self, fill_type, columns: [] = None, table_max_size=1, key_max_size=-1, replication=MINIMUM_REPLICATION) -> Promise:
        servers_req = get_servers().wait_response()
        if not servers_req.success:
            return Promise.reject(CreateTableResponse(False,CreateTableError.INTERNAL_ERROR))  
        
        return create_table(self.name, fill_type, columns, servers_req.servers, key_max_size=key_max_size, table_max_size=table_max_size, replication=replication)

    def restructure(self, columns: [], servers: [], table_max_size=1, replication=MINIMUM_REPLICATION) -> Promise:
        return restructure_table(self.name, columns, servers, replication, table_max_size)

    def drop(self) -> Promise:
        return delete_table(self.name)

    def truncate(self) -> Promise:
        return truncate_table(self.name)

    def get_structure(self) -> Promise:
        return get_table_structure(self.name)



from ayradb.core.manage_table.create_lists import create_server_list_from_dict, create_column_list
from ayradb.rest.promise import Promise
from ayradb.core.manage_table.create.response import CreateTableResponse, CreateTableError
from ayradb.core.manage_table.column import COLUMN_MAX_LENGTH, COLUMN_NAME_KEY
from ayradb.core.manage_table.types import TABLE_FIXED_LENGTH
from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.core.endpoints.endpoints import SYSTEM_ENDPOINT, TABLES_ENDPOINT
import json

ACTION_TABLE_CREATE = "table_create"
USER_TABLE = "user_table"
MINIMUM_REPLICATION = 1
DEFAULT_ESCAPE_FACTOR = 1  # 100%


def create_table(name, fill_type, user_columns:[], servers:dict, replication=1, table_max_size=1, tablet_max_size=1, key_max_size =-1, escape_factor=DEFAULT_ESCAPE_FACTOR):
    # Validate received fields:
    if replication > MINIMUM_REPLICATION > servers.__len__():
        return Promise.reject(CreateTableResponse(False,CreateTableError.NOT_ENOUGH_NODES))
    if fill_type == TABLE_FIXED_LENGTH:
        for idx in range(0,user_columns.__len__()):
            if user_columns[idx].get(COLUMN_MAX_LENGTH) is None:
                return Promise.reject(CreateTableResponse(False, CreateTableError.FIELD_MAX_LENGTH_REQUIRED))
        if key_max_size <= 0:
            return Promise.reject(CreateTableResponse(False, CreateTableError.KEY_MAX_LENGTH_REQUIRED))
    # Create request
    server_list = create_server_list_from_dict(servers)
    # Notice add 1 to escape factor to allocate the correct size
    column_list = create_column_list(user_columns, fill_type, escape_factor+1, key_max_size=key_max_size)
    table_description = {
        "table_conventional_name": name,
        "table_category": "main",
        "table_clone_name": "",
        "table_main_name": "",
        "table_tsynch_name": "",
        "table_level_for_read": 0,
        "table_level_for_write": 0,
        "table_level_for_modify": 0,
        "use_default_for_table_max_size": "yes" if table_max_size == 1 else "no",
        "use_default_for_tablet_max_size": "yes" if tablet_max_size == 1 else "no",
        "table_max_size_byte": table_max_size,  # Ignored when use_default_for_table_max_size is yes
        "tablet_max_size_byte": tablet_max_size,   # Ignored when use_default_for_tablet_max_size is yes
        "table_fill_type": fill_type,
        "record_padded_length_mod": 1000,
        "replication_factor": replication,
        "n_columns": column_list.__len__(),
        "primary_index_column": COLUMN_NAME_KEY,
        "output_records_type": "straight_row",
        "output_json_generation_mode": "standard",
        "table_strong_synchronization": "off",
        "table_allocation_servers": server_list,
        "column_descriptions": column_list,
    }

    query = {
        "action": ACTION_TABLE_CREATE
    }

    body = {
        "endpoint": TABLES_ENDPOINT,
        "table_type": USER_TABLE,
        "table_description": table_description
    }

    create_request = Request(Request.METHOD_POST, path=SYSTEM_ENDPOINT, query=query, body=json.dumps(body).encode('ascii'))
    return RestManager().submit(create_request, CreateTableResponse.from_http_response)


from ayradb.core.manage_table.delete.response import DeleteTableResponse
from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.core.endpoints.endpoints import SYSTEM_ENDPOINT

ACTION = "table_destroy"


def delete_table(table_name):
    query = {
        "action": ACTION,
        "table_name": table_name
    }
    delete_table_request = Request(Request.METHOD_POST, path=SYSTEM_ENDPOINT, query=query)
    return RestManager().submit(delete_table_request, DeleteTableResponse.from_http_response)



import json

from ayradb.core.manage_record.scan.actions import ACTION_SCAN_TABLE, SUB_ACTION_SCAN_INIT
from ayradb.core.manage_record.scan.response_scan_init import ScanInitResponse
from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.core.endpoints.endpoints import SYSTEM_ENDPOINT


def scan_init(table_name):
    # Create query
    query = {
        "action": ACTION_SCAN_TABLE,
    }
    # Create body
    body = {
        "table_name": table_name,
        "subaction": SUB_ACTION_SCAN_INIT
    }

    json_body = json.dumps(body).encode('ascii')
    scan_init_request = Request(Request.METHOD_POST, path=SYSTEM_ENDPOINT, query=query, body=json_body)
    return RestManager().submit(scan_init_request, ScanInitResponse.from_http_response)


from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name="ayradb",
    version="0.3.0-beta",
    licence="Apache-2.0",
    author="CherryData srl",
    author_email="info@cherry-data.com",
    description="AyraDB python client",
    long_description = long_description,
    long_description_content_type="text/markdown",
    url="https://www.ayradb.com",
    keywords=["AyraDB", "client", "noSQL", "database", "connector", "key-value"],
    package_dir={"": "src"},
    packages=find_packages(where='src'),
    install_requires=['dataclasses'],
)



__all__ = ['DeleteResponse', 'DeleteError']

from ayradb.rest.http.response import Response
import re
from dataclasses import dataclass


class DeleteError:
    # Request errors (local)
    INVALID_FIELD_NAME=-1
    # AyraDB errors
    TABLE_NOT_FOUND=1
    FIELD_NOT_FOUND=2
    INTERNAL_ERROR=100


ERROR_DICT = {
    "ERROR: [00405]:(00008): TFA_EEPA_019: No parameter group found implementing the requested action: probably action group value":DeleteError.TABLE_NOT_FOUND,
    "ERROR: [00017]:(00001): TFA_MDAI_016.003: the specified field does not exist": DeleteError.FIELD_NOT_FOUND,
    # Errors that shouldn't happen since they are internally managed
    "ERROR: [00016]:(00002): TFA_EEPA_018: endpoint action can't be done with this request method": DeleteError.INTERNAL_ERROR,
}


@dataclass
class DeleteResponse:

    success : bool
    error_code: int
    _error_msg: str

    def __init__(self, success, error_code,_error_msg=""):
        self.success=success
        self.error_code=error_code
        self._error_msg=_error_msg

    @staticmethod
    def from_http_response(res: Response):
        error_code = 0
        _error_msg = ""
        if res.status_code == 200:
            success = True
        else:
            # Case error returned by ayra
            success = False
            error = res.body.decode('ascii')
            # Search for server name inside error message
            match = re.search(r'^(([\s\S])+(?=(ERROR)))', error)
            if match is not None:
                # Case server name present in error
                error=error[match.end():]  # Remove server name from error
            internal_error=DeleteError.INTERNAL_ERROR
            try:
                error_code=ERROR_DICT[error]
            except KeyError:
                error_code=internal_error
            if error_code == internal_error:
                _error_msg=error
                # TODO: save in a log file

        return DeleteResponse(success, error_code, _error_msg=_error_msg)



__all__ = ['GetServersResponse', 'GetServersError']

from ayradb.rest.http.response import Response
import re
from dataclasses import dataclass


class GetServersError:
    # AyraDB errors
    INTERNAL_ERROR = 100


SEPARATOR = b'\x3b'
ERROR_DICT = {
    # Errors that shouldn't happen since they are internally managed
    "ERROR: [00016]:(00002): TFA_EEPA_018: endpoint action can't be done with this request method": GetServersError.INTERNAL_ERROR,
}


@dataclass
class GetServersResponse:

    success: bool
    servers: dict
    error_code: int
    _error_msg: str

    def __init__(self, success, error_code, servers=None, _error_msg=""):
        self.success = success
        self.error_code = error_code
        self.servers = servers
        self._error_msg = _error_msg

    @staticmethod
    def from_http_response(res: Response):
        if res.status_code == 200:
            servers = {}
            body: bytes = res.body
            splitted_body = body.split(SEPARATOR)
            # Parse body
            for cursor in range(0,splitted_body.__len__(),2):
                # splitted body is organized as [srv-name, srv-ip, srv-name, srv-ip,...]
                field_key = splitted_body[cursor]
                field_value = splitted_body[cursor+1]
                servers[field_key.decode('ascii')] = field_value.decode('ascii')
            return GetServersResponse(True, 0, servers=servers)
        else:
            # Case error returned by ayra
            _error_msg = ""
            error = res.body.decode('ascii')
            # Search for server name inside error message
            match = re.search(r'^(([\s\S])+(?=(ERROR)))', error)
            if match is not None:
                # Case server name present in error
                error = error[match.end():]  # Remove server name from error
            internal_error = GetServersError.INTERNAL_ERROR
            try:
                error_code = ERROR_DICT[error]
            except KeyError:
                error_code = internal_error
            if error_code == internal_error:
                _error_msg = error
                # TODO: save in a log file
            return GetServersResponse(False, error_code, _error_msg=_error_msg)



import json

from ayradb.core.manage_record.scan.actions import ACTION_SCAN_TABLE, SUB_ACTION_SCAN_START, SUB_ACTION_SCAN_CONTINUE
from ayradb.core.manage_record.scan.response_scan import ScanResponse, ScanError
from ayradb.core.validate import validate_field
from ayradb.rest.rest_manager import RestManager
from ayradb.rest.http.request import Request
from ayradb.rest.promise import Promise
from ayradb.core.endpoints.endpoints import SYSTEM_ENDPOINT


def scan(table_name, segment, fields=None, last_hash=None):
    # Create query
    query = {
        "action": ACTION_SCAN_TABLE,
    }
    # Create body common fields
    body = {
        "table_name": table_name,
        "subaction": SUB_ACTION_SCAN_CONTINUE,
        "last_key_hash": last_hash,
        "segment": str(segment),
    }
    if last_hash is None:
        # Less frequent (case scan start)
        body["subaction"] = SUB_ACTION_SCAN_START
        del body["last_key_hash"]
    if fields is None or fields.__len__() == 0:
        # Case scan all fields
        body["fields"] = "*"
    else:
        # Case scan only requested fields
        concat = fields[0]
        fields_valid = validate_field(concat)
        for fieldIdx in range(1, fields.__len__()):
            # Concat remaining fields
            curr_field = fields[fieldIdx]
            fields_valid = fields_valid and validate_field(curr_field)
            concat = f'{concat},{curr_field}'
        body["fields"] = concat

        if not fields_valid:
            # Case at least one invalid field
            return Promise.reject(ScanResponse(False, ScanError.INVALID_FIELD_NAME, None, None))

    # Case inserted fields are all valid or fields=="*"
    scan_start_request = \
        Request(Request.METHOD_POST, path=SYSTEM_ENDPOINT, query=query, body=json.dumps(body).encode('ascii'))
    return RestManager().submit(scan_start_request, ScanResponse.from_http_response)



from ayradb.rest.promise import Promise
from ayradb.rest.http.request import Request
from ayradb.rest.http.header import Header
from ayradb.core.endpoints.endpoints import SYSTEM_ENDPOINT
from ayradb.core.keep_alive.response import KeepAliveResponse


def generate_keep_alive_promise():
    query = {
        "action": "keep_alive"
    }

    headers = {
        Header.CONNECTION: "Keep-Alive"
    }

    req = Request(Request.METHOD_GET, path=SYSTEM_ENDPOINT, query=query, headers=headers)
    return Promise(req, KeepAliveResponse.from_http_response)



from ayradb.core.manage_table.column import COLUMN_NUMBER, COLUMN_LABEL, COLUMN_NAME_KEY, COLUMN_SQL_TYPE, SQLFieldType, \
    COLUMN_HTTP_TYPE, HTTPFieldType, COLUMN_MAX_LENGTH, COLUMN_NAME_NOSQL_VALUE
from ayradb.core.manage_table.types import TABLE_FIXED_LENGTH, TABLE_NOSQL


def create_server_list_from_list(server_names: list) -> list:
    for server_name in server_names:
        server_names.append({
            "server_name": server_name
        })
    return server_names


def create_server_list_from_dict(servers: dict) -> list:
    server_names = []
    for server_name in servers.keys():
        server_names.append({
            "server_name": server_name
        })
    return server_names


def create_column_list(user_columns: [], fill_type, escape_factor, key_max_size=-1) -> list:
    columns = [{
        COLUMN_NUMBER: 0,
        COLUMN_LABEL: COLUMN_NAME_KEY,
        COLUMN_SQL_TYPE: SQLFieldType.TEXT,
        COLUMN_HTTP_TYPE: HTTPFieldType.TEXT_HTML
    }]
    if fill_type == TABLE_FIXED_LENGTH:
        # Case fixed length: specify max key size (consider escape factor)
        true_key_max_size=key_max_size*escape_factor
        columns[0][COLUMN_MAX_LENGTH]=true_key_max_size

    if fill_type == TABLE_NOSQL:
        # Case NO_SQL table: add value column
        columns.append({
            COLUMN_NUMBER: 1,
            COLUMN_LABEL: COLUMN_NAME_NOSQL_VALUE,
            COLUMN_SQL_TYPE: SQLFieldType.TEXT,
            COLUMN_HTTP_TYPE: HTTPFieldType.TEXT_HTML
        })
    else:
        is_fixed_length= fill_type == TABLE_FIXED_LENGTH
        for cIdx in range(0,user_columns.__len__()):
            user_columns[cIdx][COLUMN_NUMBER]=cIdx+1
            if is_fixed_length:
                # Case fixed length: specify max field size (consider escape factor)
                column_true_size = user_columns[cIdx][COLUMN_MAX_LENGTH]*escape_factor
                user_columns[cIdx][COLUMN_MAX_LENGTH] = column_true_size
            if user_columns[cIdx].get(COLUMN_SQL_TYPE) is None:
                # Case sql type not specified: defaulted to text
                user_columns[cIdx][COLUMN_SQL_TYPE] = SQLFieldType.TEXT
            if user_columns[cIdx].get(COLUMN_HTTP_TYPE) is None:
                # Case http type not specified: defaulted to text/html
                user_columns[cIdx][COLUMN_HTTP_TYPE] = HTTPFieldType.TEXT_HTML
        columns.extend(user_columns)
    return columns

